/* $Id: inet_acct.c,v 1.4 1994/09/10 20:40:42 root Exp $
 *
 * Copyright (C) 1994 Lars Fenneberg
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <linux/config.h>

#include <linux/sched.h>
#include <linux/socket.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>

#include <asm/segment.h>
#include <linux/malloc.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/in.h>
#include <linux/inet.h>
#include <linux/net.h>
#include <linux/inet_acct.h>

#undef DEBUG

static struct ipacct_user_s *ip_stat[USER_HASHMAX];
static struct ipacct_weight_s *weight_head, *weight_tail;
static int default_weight = DEFAULT_WEIGHT;

/***********************************************************************/

static inline struct ipacct_weight_s *
ipacct_match_addr(unsigned long waddr)
{ 
  struct ipacct_weight_s *p;

  for (p = weight_head; p != NULL; p = p->succ) 
  {
    if (!(((p->addr) ^ waddr) & (p->mask))) break;
  }

#ifdef DEBUG
  if ((p != NULL) && (p->magic != WEIGHT_MAGIC))
  {
    printk("ipacct: weight list corrupt.\n");
    p = NULL;
  }
#endif

  return p;
}


static int
ipacct_add_addr(struct ipacct_weight_entry *wt)
{
  struct ipacct_weight_s *p = weight_head; 

  if ((wt->addr) == htonl(0x00000000)) {
    default_weight = wt->weight;
    return 0;  
  }

  cli();
  while ((p != NULL) && 
        (((p->addr) != (wt->addr)) ||
        ((p->mask) != (wt->mask)))) p = p->succ;

#ifdef DEBUG
  if ((p != NULL) && (p->magic != WEIGHT_MAGIC))
  {
    printk("ipacct: weight list corrupt.\n");
    p = NULL;
  }
#endif
  
  if (p != NULL) 
  {
    p->weight = wt->weight;
  }
  else
  {
    if ((p = (struct ipacct_weight_s *)kmalloc(sizeof(struct ipacct_weight_s),GFP_KERNEL)) == NULL) return -ENOMEM;
    
    p->magic  = WEIGHT_MAGIC;
    p->addr   = wt->addr;
    p->mask   = wt->mask;
    p->weight = wt->weight;
    p->succ   = NULL;

    if (weight_head == NULL) 
    {
    	weight_head = weight_tail = p;
    }
    else
    {
    	weight_tail->succ = p;
    	weight_tail = p;
    }

  }
  sti();

  return 0;
}

static int
ipacct_del_addr(struct ipacct_weight_entry *wt)
{
  struct ipacct_weight_s *p = weight_head;
  struct ipacct_weight_s *pred = NULL;

  cli();  

  while ((p != NULL) && 
        (((p->addr) != (wt->addr)) ||
        ((p->mask) != (wt->mask)))) 
  {
    pred = p;
    p = p->succ;
  }

#ifdef DEBUG
  if ((p != NULL) && (p->magic != WEIGHT_MAGIC))
  {
    printk("ipacct: weight list corrupt.\n");
    p = NULL;
  }
#endif

  if (p == NULL) return -EINVAL;
	
  if (pred == NULL)
  {
    weight_head = p->succ;
    if (weight_head == NULL) weight_tail = NULL;
  }
  else 
  {
    pred->succ = p->succ;
    if (pred->succ == NULL) weight_tail = pred;
  }

  sti();
  kfree_s(p,sizeof(struct ipacct_weight_s));

  return 0;
}

static inline int
ipacct_get_weight(unsigned long waddr)
{
  struct ipacct_weight_s *p;
  int w;

  cli();  
  if ((p = ipacct_match_addr(waddr)) == NULL) w = default_weight; else w = (p->weight);
  sti();
  
  return w;
}

/***********************************************************************/
 
static struct ipacct_user_s *
ipacct_find_user(uid_t user)
{ 
  struct ipacct_user_s *p = ip_stat[USER_HASHVAL(user)];

  while ((p != NULL) && ((p->uid) != user)) p = p->succ;

#ifdef DEBUG
  if ((p != NULL) && (p->magic != USER_MAGIC))
  {
    printk("ipacct: user list corrupt.\n");
    p = NULL;
  }
#endif

  return p;
}

/***********************************************************************/

int
ipacct_sent(uid_t user,unsigned long daddr, int sent)
{
  struct ipacct_user_s *p; 
  int hashval;

  hashval = USER_HASHVAL(user);

  if (sent<=0) return 0;

#ifdef DEBUG
  printk("ipacct: user %i daddr %s sent %i\n",user,in_ntoa(daddr),sent);
#endif

  cli();
  
  if ((p = ipacct_find_user(user)) != NULL) 
  {
  
    p->sent += sent*ipacct_get_weight(daddr);
    
  }
  else
  {
  
    if ((p = (struct ipacct_user_s *)kmalloc(sizeof(struct ipacct_user_s),GFP_ATOMIC)) == NULL) return -ENOMEM;
    
    p->magic  = USER_MAGIC;
    p->uid    = user;
    p->sent   = sent;
    p->recv   = 0;
   
    p->succ   = ip_stat[hashval];
    ip_stat[hashval] = p; 

  }

  sti();
  return 0;
}

int
ipacct_received(uid_t user, unsigned long daddr, int recv)
{
  struct ipacct_user_s *p; 
  int hashval;

  hashval = USER_HASHVAL(user);

  if (recv<=0) return 0;

#ifdef DEBUG
  printk("ipacct: user %i daddr %s recv %i\n",user,in_ntoa(daddr),recv);
#endif
  
  cli();
  
  if ((p = ipacct_find_user(user)) != NULL) 
  {
  
    p->recv += recv*ipacct_get_weight(daddr);
    
  }
  else
  {
  
    if ((p = (struct ipacct_user_s *)kmalloc(sizeof(struct ipacct_user_s),GFP_ATOMIC)) == NULL) return -ENOMEM;
    
    p->magic  = USER_MAGIC;
    p->uid    = user;
    p->sent   = 0;
    p->recv   = recv;

    p->succ   = ip_stat[hashval];
    ip_stat[hashval] = p; 

  }
  sti();

  return 0;
}

/***********************************************************************/

int
ipacct_get_user_info(char *buffer, char **start, off_t offset, int length)
{
	struct ipacct_user_s *p;
	int i;	
	int len=0;
	off_t pos=0;
	off_t begin=0;
  
	len+=sprintf(buffer, "uid     sent     recv\n");

	for(i = 0; i < USER_HASHMAX; i++) 
	{
	  	cli();
		p = ip_stat[i];
		while(p != NULL) 
		{
			
			len+=sprintf(buffer+len,"%-8i %-10i %-10i\n",p->uid,p->sent,p->recv);
		
			p = p->succ;
			
			pos=begin+len;
			if(pos<offset)
			{
				len=0;
				begin=pos;
			}
			if(pos>offset+length)
				break;
		}
		sti();	/* We only turn interrupts back on for a moment, but because the interrupt queues anything built up
			   before this will clear before we jump back and cli, so its not as bad as it looks */
		if(pos>offset+length)
			break;
	}
	*start=buffer+(offset-begin);
	len-=(offset-begin);
	if(len>length)
	  	len=length;
	return len;
} 

int
ipacct_get_weight_info(char *buffer, char **start, off_t offset, int length)
{
	struct ipacct_weight_s *p = weight_head;
	int len=0;
	off_t pos=0;
	off_t begin=0;
  
	len+=sprintf(buffer, "addr\t\tmask\t\tweight\n");

	cli();
	while(p != NULL) 
	{

	  len+=sprintf(buffer+len,"%08lX\t%08lX\t%-5i\n",p->addr,p->mask,p->weight);

	  p = p->succ;
			
	  pos=begin+len;
	  if(pos<offset)
	  {
	    len=0;
	    begin=pos;
	  }
	  if(pos>offset+length) break;
	}
	sti();

	if (pos<=offset+length) {

	  len+=sprintf(buffer+len,"%08lX\t%08lX\t%-5i\n",htonl(0x00000000),htonl(0xffffffff),default_weight);

	  pos=begin+len;
	  if(pos<offset)
	  {
	    len=0;
	    begin=pos;
	  }
	
	}

	*start=buffer+(offset-begin);
	len-=(offset-begin);
	if(len>length)
	  	len=length;
	return len;
} 

/***********************************************************************/

void 
ipacct_init(void)
{
  int i;

  printk(IPACCT_VERSION);
  
  for(i=0; i<USER_HASHMAX; i++) ip_stat[i] = NULL;
  weight_head = weight_tail = NULL;

}

/***********************************************************************/

int ipacct_ioctl(unsigned int cmd, void *arg)
{
	int err;
	struct ipacct_weight_entry wt;

	switch(cmd) 
	{
		case SIOCADDWT:		/* Add a weight */
		case SIOCDELWT:		/* Delete a weight */
			if (!suser())
				return -EPERM;
			err=verify_area(VERIFY_READ, arg, sizeof(struct ipacct_weight_entry));
			if (err)
				return err;
			memcpy_fromfs(&wt, arg, sizeof(struct ipacct_weight_entry));
			return (cmd == SIOCADDWT) ? ipacct_add_addr(&wt) : ipacct_del_addr(&wt);
	}

	return -EINVAL;
}
